import java.io.File;
import java.io.IOException;
import javax.sound.sampled.*;
import javax.sound.sampled.FloatControl.Type;

import java.io.ByteArrayOutputStream;
import java.nio.charset.Charset;


// TODO: ミキサーのマスターボリュームを変化させたい。Win10 では API 叩かなければならないのか。
public class Audio extends Thread {
    public static int SAMPLING_RATE = 48000;
    public static int SOUND_FREQ = 880;
    private int times;
    private int msec;
    private String file;
    private String targetName;
    private String mixerName;
    private int vol = 50;
    private AudioInputStream audio;
    
    public Audio(String targetName, String mixerName, int vol, int times, int msec) {
    	this.targetName = targetName;
    	this.mixerName = mixerName;
    	this.vol = vol;
        this.times = times;
        this.msec  = msec;
        audio = new WaveBeep(SAMPLING_RATE, SOUND_FREQ, msec/2);
    }
    
    public Audio(String targetName, String mixerName, int vol, String file) {
    	this.targetName = targetName;
    	this.mixerName = mixerName;
        this.vol = vol;
        this.file = file;
       
        try {
        	audio = AudioSystem.getAudioInputStream(new File(file));
        } catch (UnsupportedAudioFileException e) {
        	e.printStackTrace();
        } catch (IOException e) {
        	e.printStackTrace();
        }
    }
    
    public Mixer.Info getMixerInfoByName(String name) {
    	if (name.startsWith("(NO CLIP)")) name=name.substring(9);
    	Mixer.Info[] info = AudioSystem.getMixerInfo();
    	for (int i=0; i<info.length; i++) {
    		if (deMessyCode(info[i].getName()).equals(name)) return info[i];
    	}
    	return null;
    }

    public static String[] getMixerNames() {
		Mixer.Info[] mixerInfo = AudioSystem.getMixerInfo();
		String[] ret = new String[mixerInfo.length];
		for (int i=0; i<mixerInfo.length; i++) {
			ret[i] = Audio.deMessyCode(mixerInfo[i].getName()); 
			Mixer mixer = AudioSystem.getMixer(mixerInfo[i]);
			boolean existClip = false;
			for (Line.Info lineInfo : mixer.getSourceLineInfo()) {
				try {
					if (AudioSystem.getLine(lineInfo) instanceof Clip) existClip = true;
				} catch (LineUnavailableException e) {
					//e.printStackTrace();
				}
			}
			if (!existClip) ret[i] = "(NO CLIP)" + ret[i];
		}
    	return ret;
    }
    
    @Override
    public void run() {
    	Mixer.Info targetInfo = null;
    	if (!targetName.isEmpty()) {
    		targetInfo = getMixerInfoByName(targetName);
    		if (targetInfo == null) {
    			System.out.println("No device for target " + targetName);
    			return;
    		}
    	}
    	Mixer.Info mixerInfo = null;
    	if (!targetName.isEmpty()) {
    		mixerInfo = getMixerInfoByName(mixerName);
    		if (mixerInfo == null) {
    			System.out.println("No device for mixer " + mixerName);
    			return;
    		}
    	}
    	//System.out.println("Sound Start");
        try {
        	Mixer mixer = AudioSystem.getMixer(mixerInfo);
        	for (Line.Info lineInfo: mixer.getTargetLineInfo()) {
        		Line line = AudioSystem.getLine(lineInfo);
        		if (line instanceof Port) System.out.println(line + "is a part of Port");
        		System.out.println(lineInfo + " | " + line);
        		line.open();
        		masterVolume = null;
        		searchMasterVolume(line.getControls());
        		if (masterVolume !=null) masterVolume.setValue(vol/100f);
        		line.close();
        	}
            mixer.close();
        	
            Clip clipTarget = AudioSystem.getClip(targetInfo);
            clipTarget.open(audio);
            if (times > 0) {
            	System.out.println("sin wave " + msec + "(ms) * " + times);
            	for (int i=0; i<times; i++) {
            		clipTarget.setFramePosition(0);
            		clipTarget.start();
            		sleep(msec);
            	}
            } else {
            	System.out.println("Sound file \"" + file + "\"");
        		clipTarget.start();
        		sleep(clipTarget.getMicrosecondLength()/1000);
            }
            clipTarget.close();
        } catch (LineUnavailableException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (IllegalArgumentException e) {
            e.printStackTrace();
        } catch (NegativeArraySizeException e) {
        	//e.printStackTrace(); // PulseAudioClip.java:450
        }
        //System.out.println("Sound End");
    }
    
    FloatControl masterVolume = null;
    int level = 0;
    private void searchMasterVolume(Control[] ctrl) {
    	System.out.println(++level);
    	int index = 0;
    	for (Control c: ctrl) {
    		if (c instanceof CompoundControl) {
    			System.out.println("Compound " + index + ":" + level + " " + c);
    			searchMasterVolume(((CompoundControl) c).getMemberControls());
    		} else {
    			if (c.getType() == Type.VOLUME) {
    				if (masterVolume == null) masterVolume = (FloatControl)c;
    				System.out.println("volume!!!!"); 
    			}
    			if (c.getType() == Type.MASTER_GAIN) System.out.println("masterGain!!!!"); 
    			System.out.println(index + " " + level + " (" +c+") "+c.getClass());
    		}
    		index++;
    	}
    	level--;
    }

    /**
     * JDKのバグ対策（サウンドデバイス名のエンコード間違いを直す）
     * https://qiita.com/TakaW/items/023b06cd1ba18ca62047
     * https://bugs.java.com/bugdatabase/view_bug.do?bug_id=JDK-8177951 
     */
    private static final byte HIGH_BIT = (byte) 0b11000000;
    public static String deMessyCode(String messyCode) {
        ByteArrayOutputStream buf = new ByteArrayOutputStream(messyCode.length());
        byte[] originalBytes = messyCode.getBytes(Charset.forName("UTF-8"));
        String systemEnc = System.getProperty("file.encoding");
        //System.out.println(systemEnc + (systemEnc.equalsIgnoreCase("MS932")));

        for (int i = 0; i < originalBytes.length; i++) {
            if ((byte) (originalBytes[i] & HIGH_BIT) == HIGH_BIT) {
            	if (i==originalBytes.length-2 && systemEnc.equalsIgnoreCase("MS932")) break;
            	// SJIS の泣き別れが起きないよう、最後の1文字は書き込まない
                buf.write((originalBytes[i] << 6) | (originalBytes[i + 1] << 2 >>> 2));
                // DELETE 0b110000 and move 6, then delete the 0b10 prefix of the second byte.
                i++;
            } else {
                buf.write(originalBytes[i]);
            }
        }
        return new String(buf.toByteArray(), Charset.forName(systemEnc));
    }
}
